from django.core.serializers.json import DjangoJSONEncoder
from django.db import models

from django_seriously.utils.forms import PydanticJSONFormField
from django_seriously.utils.pydantic import PydanticMixin


class ValidatedJSONField(PydanticMixin, models.JSONField):
    """
    Model field that validates json data structure according to specified
    pydantic model. Data remains in generic python objects, while pydantic
    is only used for validation on save.
    """

    def __init__(self, structure, **kwargs):
        self.structure = structure
        if structure in [str, bytes]:
            raise ValueError("ValidatedJSONField requires non-trivial structure")
        kwargs.setdefault("encoder", DjangoJSONEncoder)
        self.json_loads = kwargs.pop("json_loads", None)
        super().__init__(**kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        kwargs["structure"] = None
        return name, path, args, kwargs

    def validate(self, value, model_instance):
        super(models.JSONField, self).validate(value, model_instance)
        self._loads(value)


class PydanticJSONField(PydanticMixin, models.JSONField):
    """ """

    def __init__(self, structure, **kwargs):
        self.structure = structure
        if structure in [str, bytes]:
            raise ValueError("PydanticJSONField requires non-trivial structure")
        kwargs.setdefault("encoder", DjangoJSONEncoder)
        self.json_loads = kwargs.pop("json_loads", None)
        self.json_dumps = kwargs.pop("json_dumps", None)
        self._initial = None
        super().__init__(**kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        kwargs["structure"] = None
        return name, path, args, kwargs

    def validate(self, value, model_instance):
        super(models.JSONField, self).validate(value, model_instance)
        # this is somewhat stupid, but pydantic only validates on load and
        # thus validation errors can only be caught with this extra step.
        # try to be a bit smarter by skipping loading when nothing changed
        dumped_value = self._dumps(value)
        if dumped_value != self._initial:
            self._loads(dumped_value)

    def from_db_value(self, value, expression, connection):
        if value is None:
            return value
        self._initial = value
        return self._loads(value)

    def to_python(self, value):
        if value is None:
            return value
        return self._loads(value)

    def get_prep_value(self, value):
        if value is None:
            return value
        return self._dumps(value)

    def formfield(self, **kwargs):
        return super().formfield(
            **{
                "form_class": PydanticJSONFormField,
                "structure": self.structure,
                "encoder": self.encoder,
                "decoder": self.decoder,
                "json_loads": self.json_loads,
                "json_dumps": self.json_dumps,
                **kwargs,
            }
        )



import pytest
from django.core import exceptions
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models

from django_seriously.utils.models import BaseModel


class TestModel(BaseModel):
    field = models.IntegerField(
        validators=[MinValueValidator(0), MaxValueValidator(100)]
    )


@pytest.mark.django_db
def test_base_model_derivation():
    inst = TestModel.objects.create(field=1)
    assert inst.created_at
    assert inst.updated_at
    assert inst.field == 1


@pytest.mark.django_db
def test_base_model_clean_check():
    """forced full_clean makes this raise. otherwise validation would not run"""
    with pytest.raises(exceptions.ValidationError):
        TestModel.objects.create(field=102)



# Generated by Django 4.1.3 on 2022-12-18 15:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("authtoken", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="token",
            name="last_seen_at",
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="token",
            name="scopes",
            field=models.CharField(
                blank=True,
                help_text="comma-separated list of scopes. choices are n/a.",
                max_length=50,
            ),
        ),
    ]



from typing import Any, Dict, List, Set

from django.utils.module_loading import import_string


def perform_import(val):
    if val is None:
        return None
    elif isinstance(val, str):
        return import_string(val)
    elif isinstance(val, (list, tuple)):
        return [import_string(item) for item in val]
    return val


class AppSettings:
    """
    Reusable settings container that handles import strings, defaults,
    reloading, and lazy evaluation.
    This class is shamelessly recycled from DRF's APISettings without
    introducing a dependency on DRF.
    """

    def __init__(
        self,
        user_settings: Dict[str, Any],
        defaults: Dict[str, Any],
        import_strings: List[str],
    ):
        self.user_settings = user_settings
        self.defaults = defaults
        self.import_strings = import_strings
        self._cached_attrs: Set[str] = set()

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError("Invalid API setting: '%s'" % attr)

        try:
            # Check if present in user settings
            val = self.user_settings[attr]
        except KeyError:
            # Fall back to defaults
            val = self.defaults[attr]

        # Coerce import strings into classes
        if attr in self.import_strings:
            val = perform_import(val)

        # Cache the result
        self._cached_attrs.add(attr)
        setattr(self, attr, val)
        return val

    def reload(self):
        for attr in self._cached_attrs:
            delattr(self, attr)
        self._cached_attrs.clear()
        if hasattr(self, "_user_settings"):
            delattr(self, "_user_settings")



from datetime import datetime

import pytest
from django.core import exceptions
from django.utils import timezone
from pydantic import BaseModel as PydanticBaseModel

from django_seriously.utils.fields import PydanticJSONField
from django_seriously.utils.models import BaseModel


class X(PydanticBaseModel):
    a: int
    b: str
    c: datetime


class PydanticFieldTestModel(BaseModel):
    field = PydanticJSONField(structure=X)
    field_list = PydanticJSONField(structure=list[X])


X_EXAMPLE_DICT = {"a": 1, "b": "foo", "c": timezone.now()}


@pytest.mark.parametrize("field_input", [X_EXAMPLE_DICT, X(**X_EXAMPLE_DICT)])
@pytest.mark.django_db
def test_pydantic_field_creation(field_input):
    instance = PydanticFieldTestModel.objects.create(
        field=field_input,
        field_list=[field_input],
    )
    assert isinstance(instance.field, X)
    assert isinstance(instance.field_list[0], X)
    instance_retrieved = PydanticFieldTestModel.objects.get()
    assert isinstance(instance_retrieved.field, X)
    assert isinstance(instance_retrieved.field_list[0], X)


@pytest.mark.django_db
def test_pydantic_field_invalid_creation():
    with pytest.raises(exceptions.ValidationError):
        PydanticFieldTestModel.objects.create(field=1, field_list=[X_EXAMPLE_DICT])
    with pytest.raises(exceptions.ValidationError):
        PydanticFieldTestModel.objects.create(field=X_EXAMPLE_DICT, field_list=[1])


@pytest.mark.django_db
def test_pydantic_field_invalid_modification():
    instance = PydanticFieldTestModel.objects.create(
        field=X_EXAMPLE_DICT, field_list=[X_EXAMPLE_DICT]
    )

    instance.field.a = 2
    instance.save()

    instance.field.a = "NaN"
    with pytest.raises(exceptions.ValidationError):
        instance.save()

    # previous validation should make this never fail
    instance.refresh_from_db()

    instance.field_list = [{"invalid": "invalid"}]
    with pytest.raises(exceptions.ValidationError):
        instance.save()



import os

import django
import pytest
from django.core import management


def pytest_configure(config):
    from django.conf import settings

    base_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

    settings.configure(
        DEBUG_PROPAGATE_EXCEPTIONS=True,
        DATABASES={
            "default": {"ENGINE": "django.db.backends.sqlite3", "NAME": ":memory:"}
        },
        SITE_ID=1,
        SECRET_KEY="not very secret in tests",
        USE_I18N=True,
        LANGUAGES=[
            ("de-de", "German"),
            ("en-us", "English"),
        ],
        LOCALE_PATHS=[base_dir + "/locale/"],
        STATIC_URL="/static/",
        ROOT_URLCONF="tests.urls",
        TEMPLATES=[
            {
                "BACKEND": "django.template.backends.django.DjangoTemplates",
                "DIRS": [],
                "APP_DIRS": True,
                "OPTIONS": {
                    "context_processors": [
                        "django.template.context_processors.debug",
                    ],
                },
            },
        ],
        MIDDLEWARE=(
            "django.contrib.sessions.middleware.SessionMiddleware",
            "django.middleware.common.CommonMiddleware",
            "django.contrib.auth.middleware.AuthenticationMiddleware",
            "django.middleware.locale.LocaleMiddleware",
        ),
        INSTALLED_APPS=(
            "django.contrib.auth",
            "django.contrib.contenttypes",
            "django.contrib.sessions",
            "django.contrib.sites",
            "django.contrib.messages",
            "django.contrib.staticfiles",
            "rest_framework",
            "django_seriously.authtoken",
            "tests",
        ),
        PASSWORD_HASHERS=(
            "django.contrib.auth.hashers.SHA1PasswordHasher",
            "django.contrib.auth.hashers.PBKDF2PasswordHasher",
            "django.contrib.auth.hashers.PBKDF2SHA1PasswordHasher",
            "django.contrib.auth.hashers.BCryptPasswordHasher",
            "django.contrib.auth.hashers.MD5PasswordHasher",
            "django.contrib.auth.hashers.CryptPasswordHasher",
        ),
        DEFAULT_AUTO_FIELD="django.db.models.AutoField",
        SILENCED_SYSTEM_CHECKS=[
            "rest_framework.W001",
            "fields.E210",
            "security.W001",
            "security.W002",
            "security.W003",
            "security.W009",
            "security.W012",
        ],
    )

    django.setup()
    # For whatever reason this works locally without an issue.
    # on TravisCI content_type table is missing in the sqlite db as
    # if no migration ran, but then why does it work locally?!
    management.call_command("migrate")


@pytest.fixture()
def no_warnings(capsys):
    """make sure test emits no warnings"""
    yield capsys
    captured = capsys.readouterr()
    assert not captured.out
    assert not captured.err



from django.core.exceptions import ValidationError
from django.forms import JSONField
from django.forms.fields import InvalidJSONInput, JSONString

from django_seriously.utils.pydantic import PydanticMixin


class PydanticJSONFormField(PydanticMixin, JSONField):
    def __init__(
        self,
        structure,
        encoder=None,
        decoder=None,
        json_loads=None,
        json_dumps=None,
        **kwargs,
    ):
        self.structure = structure
        self.json_loads = json_loads
        self.json_dumps = json_dumps
        super().__init__(encoder=encoder, decoder=decoder, **kwargs)

    def to_python(self, value):
        if self.disabled:
            return value
        if value in self.empty_values:
            return None
        elif isinstance(value, (list, dict, int, float, JSONString)):
            return value
        try:
            converted = self._loads(value)
        except ValidationError:
            raise ValidationError(
                self.error_messages["invalid"],
                code="invalid",
                params={"value": value},
            )
        if isinstance(converted, str):
            return JSONString(converted)
        else:
            return converted

    def bound_data(self, data, initial):
        if self.disabled:
            return initial
        if data is None:
            return None
        try:
            return self._loads(data)
        except ValidationError:
            return InvalidJSONInput(data)

    def prepare_value(self, value):
        if isinstance(value, InvalidJSONInput):
            return value
        return self._dumps(value)

    def has_changed(self, initial, data):
        if super().has_changed(initial, data):
            return True
        # For purposes of seeing whether something has changed, True isn't the
        # same as 1 and the order of keys doesn't matter.
        return self._dumps(initial) != self._dumps(self.to_python(data))



import base64
import uuid
from typing import TYPE_CHECKING, Optional

from django.contrib.auth.hashers import check_password
from django.core.exceptions import ImproperlyConfigured
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from rest_framework import exceptions
from rest_framework.authentication import BaseAuthentication, get_authorization_header
from rest_framework.permissions import BasePermission

from django_seriously.settings import seriously_settings

if TYPE_CHECKING:
    from django_seriously.authtoken.models import Token


class TokenAuthentication(BaseAuthentication):
    """
    Authentication method that works in tandem with Token.
    """

    keyword = "Bearer"
    model = None

    def get_model(self):
        if self.model is not None:
            return self.model
        return seriously_settings.AUTH_TOKEN_MODEL

    def authenticate(self, request):
        auth = get_authorization_header(request).split()

        if not auth or auth[0].lower() != self.keyword.lower().encode():
            return None

        if len(auth) == 1:
            msg = _("Invalid token header. No credentials provided.")
            raise exceptions.AuthenticationFailed(msg)
        elif len(auth) > 2:
            msg = _("Invalid token header. Token string should not contain spaces.")
            raise exceptions.AuthenticationFailed(msg)

        try:
            token_str = auth[1].decode()
        except UnicodeError:
            msg = _(
                "Invalid token header. Token string should not contain invalid characters."
            )
            raise exceptions.AuthenticationFailed(msg)

        return self.authenticate_credentials(token_str)

    def authenticate_credentials(self, token_str):
        model = self.get_model()

        try:
            token_bytes = base64.urlsafe_b64decode(token_str)
            if len(token_bytes) != 32:
                raise ValueError()
            token_id = uuid.UUID(bytes=token_bytes[:16], version=4)
            raw_token = token_bytes[16:]
        except ValueError:
            raise exceptions.AuthenticationFailed(_("Invalid token."))

        try:
            token: "Token" = model.objects.select_related("user").get(id=token_id)
        except model.DoesNotExist:
            raise exceptions.AuthenticationFailed(_("Invalid token."))

        if not check_password(password=raw_token, encoded=token.key):  # type: ignore
            raise exceptions.AuthenticationFailed(_("Invalid token."))

        if not self.check_expiration(token):
            raise exceptions.AuthenticationFailed(_("Invalid token."))

        if not token.user.is_active:
            raise exceptions.AuthenticationFailed(_("User inactive or deleted."))

        token.last_seen_at = timezone.now()
        token.save(update_fields=["last_seen_at"])

        if seriously_settings.CHECK_PASSWORD_REHASH(token.key):
            token.key = seriously_settings.MAKE_PASSWORD(raw_token)
            token.save(update_fields=["key"])

        return token.user, token

    def check_expiration(self, token: "Token"):
        """user method that handles expired tokens"""
        return True

    def authenticate_header(self, request):
        return self.keyword


class TokenHasScope(BasePermission):
    """Derived from django-oauth-toolkit's TokenHasScope"""

    def has_permission(self, request, view):
        token: Optional["Token"] = request.auth

        if not token:
            return False

        if hasattr(token, "scopes"):
            required_scopes = self.get_scopes(request, view)
            return all(r in token.scope_list for r in required_scopes)

        assert False, (
            "TokenHasScope requires the`django_seriously.authtoken.TokenAuthentication` "
            "authentication class to be used."
        )

    def get_scopes(self, request, view):
        try:
            return getattr(view, "required_scopes")
        except AttributeError:
            raise ImproperlyConfigured(
                "TokenHasScope requires the view to define the required_scopes attribute"
            )



import abc
from typing import Any, Generic, Optional, Type, TypeVar, Union

from django.contrib import messages
from django.contrib.auth.mixins import AccessMixin
from django.db import models, transaction
from django.http.response import HttpResponse, JsonResponse
from django.urls.base import reverse
from django.urls.conf import include, path
from django.utils.functional import Promise
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from django.views.generic.base import View

_T = TypeVar("_T", bound=models.Model)


class AdminRequiredMixin(AccessMixin):
    """Verify that the current user is authenticated and is admin"""

    def dispatch(self, request, *args, **kwargs):
        if not request.user.is_authenticated or not request.user.is_staff:
            return self.handle_no_permission()
        return super().dispatch(request, *args, **kwargs)  # type: ignore


class AdminItemAction(AdminRequiredMixin, View, Generic[_T], metaclass=abc.ABCMeta):
    """
    convenience view for having a context sensitive button per item in the admin list view.
    """

    _registry: list[Type["AdminItemAction"]] = []
    model_cls: Type[_T]
    successful_message = _("Action completed successfully")
    error_message = _("Action failed: {}")
    actions = [
        ("nop", _("No operation")),
    ]

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)  # type: ignore
        cls._registry.append(cls)

    def post(self, request, id: str, action: str):
        try:
            # try to obtain referenced model object
            obj = self.model_cls.objects.get(id=id)
            # see if object is still actionable (might have changed in the meantime)
            with transaction.atomic():
                if not self.is_actionable(obj, action):
                    raise ValueError("not actionable anymore")
                res = self.perform_action(obj, action)
            messages.success(request, self.successful_message)
            if res is None:
                return HttpResponse(status=204)
            else:
                return JsonResponse(res, safe=False)
        except Exception as e:
            messages.error(request, self.error_message.format(e))
            return HttpResponse(status=400)

    @abc.abstractmethod
    def perform_action(self, obj: _T, action: str) -> Any:
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def is_actionable(cls, obj: _T, action: str) -> bool:
        """does the action apply to this particular object"""
        return True

    @classmethod
    def extra_javascript(cls):
        return ""

    @classmethod
    def action_markup(cls, obj: _T):
        return mark_safe(
            " ".join([cls._action(obj, action, label) for action, label in cls.actions])
        )

    @classmethod
    def _action(cls, obj: _T, action: str, label: Union[str, Promise]):
        """template rendering of action"""
        if not cls.is_actionable(obj, action):
            return ""

        return format_html(
            """
            <a class="button" href="#" style="white-space: nowrap" onclick="
                if (!confirm('Are you sure performing this action?')) {{
                    return false;
                }}
                fetch('{}', {{
                    method: 'POST',
                    credentials: 'same-origin',
                    headers: {{
                        'X-CSRFTOKEN': document.querySelector('#changelist-form > input[type=hidden]').value
                    }}
                }})
                .then(async (resp) => {{
                    {}
                    location.reload(true);
                }})
                .catch(err => alert('request failed. please try again.'));
                return false;">{}</a>
            """,  # noqa: E501
            reverse(cls.__name__.lower(), args=[obj.pk, action]),
            cls.extra_javascript(),
            label,
        )

    @classmethod
    def _path(cls):
        """returns a urlpattern for registration in settings"""
        return path(
            route=(
                f"{cls.model_cls._meta.app_label}/{cls.model_cls.__name__.lower()}/"
                f"<uuid:id>/{cls.__name__.lower()}/<str:action>/"
            ),
            view=cls.as_view(),
            name=cls.__name__.lower(),
        )

    @classmethod
    def urls(cls):
        return include([item._path() for item in cls._registry])


def admin_navigation_link(entity: models.Model, label: Optional[str] = None) -> str:
    if not entity:
        return ""

    url = reverse(
        f"admin:{entity._meta.app_label}_{entity._meta.model_name}_change",
        args=[entity.pk],
    )
    if not label:
        label = str(entity)
    return format_html(f'<a href="{url}">{label}</a>')



from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from django_seriously.settings import seriously_settings
from django_seriously.utils.models import BaseModel


class Token(BaseModel):
    name = models.CharField(max_length=25, blank=True)
    key = models.CharField(_("Key"), max_length=128)
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name="auth_tokens",
        on_delete=models.CASCADE,
    )
    scopes = models.CharField(
        blank=True,
        max_length=50,
        help_text=(
            f"comma-separated list of scopes. choices are "
            f"{','.join(seriously_settings.AUTH_TOKEN_SCOPES) or 'n/a'}."
        ),
    )
    last_seen_at = models.DateTimeField(blank=True, null=True)

    @cached_property
    def scope_list(self):
        return self.scopes.split(",")

    def clean(self) -> None:
        super().clean()
        if not self.scopes:
            scopes = []
        elif isinstance(self.scopes, (list, tuple)):
            scopes = self.scopes
        elif isinstance(self.scopes, str):
            scopes = self.scopes.split(",")
        else:
            raise ValidationError({"scopes": "invalid scopes input"})

        valid_scopes = seriously_settings.AUTH_TOKEN_SCOPES
        if valid_scopes and not all([s in valid_scopes for s in scopes]):
            raise ValidationError(
                {"scopes": f"invalid scope choices. valid choices are: {valid_scopes}"}
            )
        self.scopes = ",".join(scopes)

    def __str__(self):
        return f"{self.name} ({self.user})"

    class Meta:
        abstract = "django_seriously.authtoken" not in settings.INSTALLED_APPS



import json
from typing import Any

from django import VERSION as DJANGO_VERSION
from django.core import exceptions
from pydantic import BaseModel, ValidationError, parse_obj_as, parse_raw_as


def _build_container(structure, *, json_loads=None, json_dumps=None):
    loads_callable = json_loads or json.loads
    dumps_callable = json_dumps or json.dumps

    class Tmp(BaseModel):
        __root__: structure  # type: ignore

        class Config:
            json_loads = loads_callable
            json_dumps = dumps_callable

    return Tmp


def _is_pydantic(value: Any) -> bool:
    try:
        return isinstance(value, BaseModel)
    except TypeError:
        return False


def pydantic_dumps(structure, value: Any, *, json_dumps=None, encoder=None) -> str:
    try:
        if _is_pydantic(value) and not json_dumps:
            obj = value
        else:
            obj = _build_container(structure, json_dumps=json_dumps)(__root__=value)

        if DJANGO_VERSION < (4, 2):
            return obj.json()
        else:
            obj = obj.dict()
            return obj["__root__"] if "__root__" in obj else obj
    except ValidationError as e:
        raise exceptions.ValidationError(f"Invalid type structure for {structure}: {e}")


def pydantic_loads(structure, value: Any, *, json_loads=None, decoder=None) -> Any:
    try:
        if json_loads:
            # enclose structure in thin wrapper to inject custom json_loads
            structure = _build_container(structure, json_loads=json_loads)
        if isinstance(value, (str, bytes)):
            result = parse_raw_as(structure, value)
        else:
            result = parse_obj_as(structure, value)
        # unpack added wrapper
        return result.__root__ if json_loads else result
    except ValidationError as e:
        raise exceptions.ValidationError(f"Invalid type structure for {structure}: {e}")


class PydanticMixin:
    def _dumps(self, value):
        return pydantic_dumps(
            self.structure, value, json_dumps=self.json_dumps, encoder=self.encoder
        )

    def _loads(self, value):
        return pydantic_loads(
            self.structure, value, json_loads=self.json_loads, decoder=self.decoder
        )



from django.contrib.auth.base_user import AbstractBaseUser
from django.contrib.auth.models import PermissionsMixin
from django.contrib.auth.models import UserManager as DjangoUserManager
from django.core.mail import send_mail
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _


class UserManager(DjangoUserManager):
    """Adaptation for minimaluser with minor changes to the Django version"""

    def _create_user(self, email, password, **extra_fields):
        email = self.normalize_email(email).lower()
        user = self.model(email=email, **extra_fields)
        user.set_password(password)
        user.save(using=self._db)
        return user

    def create_user(self, email, password=None, **extra_fields):
        extra_fields.setdefault("is_staff", False)
        extra_fields.setdefault("is_superuser", False)
        return self._create_user(email, password, **extra_fields)

    def create_superuser(self, email, password=None, **extra_fields):
        extra_fields.setdefault("is_staff", True)
        extra_fields.setdefault("is_superuser", True)

        if extra_fields.get("is_staff") is not True:
            raise ValueError("Superuser must have is_staff=True.")
        if extra_fields.get("is_superuser") is not True:
            raise ValueError("Superuser must have is_superuser=True.")

        return self._create_user(email, password, **extra_fields)


class MinimalAbstractUser(AbstractBaseUser, PermissionsMixin):
    """
    An abstract base class implementing a fully featured User model with
    admin-compliant permissions. Slight adaptation from Django version.

    Subclass (and optionally adapt) this model in your main app.
    """

    email = models.EmailField(
        _("email address"),
        unique=True,
        error_messages={
            "unique": _("A user with that email already exists."),
        },
    )
    is_staff = models.BooleanField(
        _("staff status"),
        default=False,
        help_text=_("Designates whether the user can log into this admin site."),
    )
    is_active = models.BooleanField(
        _("active"),
        default=True,
        help_text=_(
            "Designates whether this user should be treated as active. "
            "Unselect this instead of deleting accounts."
        ),
    )
    date_joined = models.DateTimeField(_("date joined"), default=timezone.now)

    objects = UserManager()

    EMAIL_FIELD = "email"
    USERNAME_FIELD = "email"
    REQUIRED_FIELDS: list[str] = []

    class Meta:
        verbose_name = _("user")
        verbose_name_plural = _("users")
        abstract = True

    def clean(self):
        super().clean()
        self.email = self.email.lower()

    def email_user(self, subject, message, from_email=None, **kwargs):
        """Send an email to this user."""
        send_mail(subject, message, from_email, [self.email], **kwargs)



from django import forms
from django.contrib.auth import password_validation
from django.contrib.auth.admin import UserAdmin as DjangoUserAdmin
from django.contrib.auth.forms import UserCreationForm as DjangoUserCreationForm
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _


class UserCreationForm(DjangoUserCreationForm):
    password1 = forms.CharField(
        label=_("Password"),
        strip=False,
        required=False,
        widget=forms.PasswordInput(attrs={"autocomplete": "new-password"}),
        help_text=password_validation.password_validators_help_text_html(),
    )
    password2 = forms.CharField(
        label=_("Password confirmation"),
        widget=forms.PasswordInput(attrs={"autocomplete": "new-password"}),
        strip=False,
        required=False,
        help_text=_("Enter the same password as before, for verification."),
    )
    no_password = forms.BooleanField(
        label=_("Create user without password"),
        required=False,
        help_text=_("Sets an invalid password"),
    )

    def _post_clean(self):
        super()._post_clean()  # type: ignore
        # Validate the password after self.instance is updated with form data
        # by super().
        password = self.cleaned_data.get("password2")
        if password:
            try:
                password_validation.validate_password(password, self.instance)
            except ValidationError as error:
                self.add_error("password2", error)
        elif self.cleaned_data.get("no_password"):
            if self.cleaned_data.get("password1") or self.cleaned_data.get("password2"):
                self.add_error(
                    "password1", _("Cannot set password when no password is requested")
                )
                self.add_error(
                    "password2", _("Cannot set password when no password is requested")
                )
        else:
            for f in ["no_password", "password1", "password2"]:
                self.add_error(f, _("Either password or flag must be set"))

    def save(self, commit=True):
        user = super().save(commit=False)
        if self.cleaned_data["no_password"]:
            user.set_unusable_password()
        else:
            user.set_password(self.cleaned_data["password1"])
        if commit:
            user.save()
        return user


class MinimalUserAdmin(DjangoUserAdmin):
    add_form = UserCreationForm
    fieldsets = (
        (None, {"fields": ("id", "email", "password")}),
        (
            _("Permissions"),
            {
                "fields": (
                    "is_active",
                    "is_staff",
                    "is_superuser",
                    "groups",
                    "user_permissions",
                ),
            },
        ),
        (_("Important dates"), {"fields": ("last_login", "date_joined")}),
    )
    add_fieldsets = (
        (
            None,
            {
                "classes": ("wide",),
                "fields": ("email", "password1", "password2", "no_password"),
            },
        ),
    )
    readonly_fields = ("id",)
    list_display = ("email", "last_login", "is_staff", "is_superuser")
    search_fields = ("email",)
    ordering = ("email",)



from typing import Any, Dict

from django.conf import settings

from django_seriously.utils.settings import AppSettings

DEFAULTS: Dict[str, Any] = {
    "AUTH_TOKEN_SCOPES": [],
    "AUTH_TOKEN_MODEL": "django_seriously.authtoken.models.Token",
    "MAKE_PASSWORD": "django_seriously.authtoken.utils.make_password",
    "CHECK_PASSWORD_REHASH": "django_seriously.authtoken.utils.check_password_rehash",
}

IMPORT_STRINGS = ["AUTH_TOKEN_MODEL", "MAKE_PASSWORD", "CHECK_PASSWORD_REHASH"]

seriously_settings = AppSettings(
    user_settings=getattr(settings, "SERIOUSLY_SETTINGS", {}),
    defaults=DEFAULTS,
    import_strings=IMPORT_STRINGS,
)



from drf_spectacular.extensions import OpenApiSerializerFieldExtension
from drf_spectacular.plumbing import follow_field_source


class PydanticJsonFieldExtensions(OpenApiSerializerFieldExtension):
    """
    extension class for drf-spectacular that hooks into JSONField
    parsing and injects `structure`'s schema instead of using a
    catch-all object.
    Simply import this file into your project to load the extension
    """

    target_class = "rest_framework.fields.JSONField"

    def map_serializer_field(self, auto_schema, direction):
        if hasattr(self.target.parent, "Meta"):
            model = self.target.parent.Meta.model
            source = self.target.source.split(".")
            model_field = follow_field_source(model, source)

            if hasattr(model_field, "structure"):
                # let pydantic generate as JSON Schema
                return model_field.structure.schema()

        return auto_schema._map_serializer_field(
            self.target, direction, bypass_extensions=True
        )



from unittest import mock

import pytest
from django.contrib.auth.hashers import get_hasher
from django.contrib.auth.models import User
from django.urls import path
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.test import APIClient
from rest_framework.views import APIView

from django_seriously.authtoken.authentication import TokenAuthentication, TokenHasScope
from django_seriously.authtoken.models import Token
from django_seriously.authtoken.utils import TokenContainer, generate_token


class TestAPIView(APIView):
    permission_classes = [IsAuthenticated]
    authentication_classes = [TokenAuthentication]

    def get(self, request):
        return Response("ok")


class TestAPIScopedView(APIView):
    permission_classes = [TokenHasScope]
    authentication_classes = [TokenAuthentication]
    required_scopes = ["test-scope1"]

    def get(self, request):
        return Response("ok")


urlpatterns = [
    path("u/", TestAPIView.as_view()),
    path("s/", TestAPIScopedView.as_view()),
]


def gen_token() -> tuple[Token, TokenContainer]:
    token_container = generate_token()
    token = Token.objects.create(
        id=token_container.id,
        key=token_container.key,
        user=User.objects.create_user("test@example.com"),
        name="test",
    )
    return token, token_container


@pytest.mark.urls(__name__)
@pytest.mark.django_db
def test_token_auth():
    token, token_container = gen_token()

    response = APIClient().get("/u/")
    assert response.status_code == 401

    response = APIClient().get(
        "/u/", HTTP_AUTHORIZATION=f"Bearer {token_container.encoded_bearer}"
    )
    assert response.status_code == 200


@pytest.mark.urls(__name__)
@pytest.mark.django_db
@mock.patch(
    "django_seriously.settings.seriously_settings.AUTH_TOKEN_SCOPES",
    ["test-scope1", "test-scope2"],
)
def test_scoped_token_auth():
    token, token_container = gen_token()

    response = APIClient().get("/s/")
    assert response.status_code == 401

    # token is missing scope
    response = APIClient().get(
        "/s/", HTTP_AUTHORIZATION=f"Bearer {token_container.encoded_bearer}"
    )
    assert response.status_code == 403

    token.scopes = "test-scope1"
    token.save()

    # token has scope
    response = APIClient().get(
        "/s/", HTTP_AUTHORIZATION=f"Bearer {token_container.encoded_bearer}"
    )
    assert response.status_code == 200


@pytest.mark.urls(__name__)
@pytest.mark.django_db
def test_token_rehash():
    token, token_container = gen_token()

    assert (
        APIClient()
        .get("/u/", HTTP_AUTHORIZATION=f"Bearer {token_container.encoded_bearer}")
        .status_code
        == 200
    )

    token.refresh_from_db()
    saved_key_original = token.key

    def make_new_password(password) -> str:
        hasher = get_hasher("pbkdf2_sha256")
        return hasher.encode(password, hasher.salt(), iterations=5_000)  # type: ignore

    def check_new_password_rehash(raw_password: str) -> bool:
        return not raw_password.startswith("pbkdf2_sha256$5000$")

    with mock.patch(
        "django_seriously.settings.seriously_settings.CHECK_PASSWORD_REHASH",
        check_new_password_rehash,
    ), mock.patch(
        "django_seriously.settings.seriously_settings.MAKE_PASSWORD", make_new_password
    ):
        assert (
            APIClient()
            .get("/u/", HTTP_AUTHORIZATION=f"Bearer {token_container.encoded_bearer}")
            .status_code
            == 200
        )

        token.refresh_from_db()
        saved_key_rehashed = token.key

        assert (
            APIClient()
            .get("/u/", HTTP_AUTHORIZATION=f"Bearer {token_container.encoded_bearer}")
            .status_code
            == 200
        )

        token.refresh_from_db()
        saved_key_rehashed2 = token.key

    assert saved_key_original.startswith("pbkdf2_sha256$1000$")
    assert saved_key_rehashed.startswith("pbkdf2_sha256$5000$")
    # no change in method, so nothing is supposed to change
    assert saved_key_rehashed == saved_key_rehashed2



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import re
import shutil
import sys

from setuptools import setup

name = "django-seriously"
package = "django_seriously"
description = (
    "Opinionated collection of Django and DRF tools that came in handy time and again."
)
url = "https://github.com/tfranzel/django-seriously"
author = "T. Franzel"
author_email = "tfranzel@gmail.com"
license = "BSD"


with open("README.rst") as readme:
    long_description = readme.read()

with open("requirements/base.txt") as fh:
    requirements = [r for r in fh.read().split("\n") if not r.startswith("#")]


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(os.path.join(package, "__init__.py")).read()
    return re.search("^__version__ = ['\"]([^'\"]+)['\"]", init_py, re.MULTILINE).group(
        1
    )


def get_packages(package):
    """
    Return root package and all sub-packages.
    """
    return [
        dirpath
        for dirpath, dirnames, filenames in os.walk(package)
        if os.path.exists(os.path.join(dirpath, "__init__.py"))
    ]


version = get_version(package)


if sys.argv[-1] == "publish":
    if os.system("pip freeze | grep twine"):
        print("twine not installed.\nUse `pip install twine`.\nExiting.")
        sys.exit(1)
    os.system("python setup.py sdist bdist_wheel")
    if os.system("twine check dist/*"):
        print("twine check failed. Packages might be outdated.")
        print("Try using `pip install -U twine wheel`.\nExiting.")
        sys.exit(1)
    if os.system("twine upload dist/*"):
        print("failed to upload package")
        sys.exit(1)
    if os.environ.get("CI"):
        os.system("git config user.name github-actions")
        os.system("git config user.email github-actions@github.com")
    os.system(f"git tag -a {version} -m 'version {version}'")
    if os.system("git push --tags"):
        print("failed pushing release tag")
        sys.exit(1)
    shutil.rmtree("dist")
    shutil.rmtree("build")
    shutil.rmtree("django_seriously.egg-info")
    sys.exit()


setup(
    name=name,
    version=version,
    url=url,
    license=license,
    description=description,
    long_description=long_description,
    long_description_content_type="text/x-rst",
    author=author,
    author_email=author_email,
    packages=get_packages(package),
    include_package_data=True,
    python_requires=">=3.6",
    install_requires=requirements,
    extras_require={},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.0",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Internet :: WWW/HTTP",
    ],
    project_urls={
        "Source": "https://github.com/tfranzel/django-seriously",
    },
)



from django import forms
from django.contrib.auth.forms import ReadOnlyPasswordHashField
from django.utils.translation import gettext_lazy as _


class TokenChangeForm(forms.ModelForm):
    key = ReadOnlyPasswordHashField(
        label=_("Key"),
        help_text=_("Raw keys are not stored, so there is no way to see this key"),
    )



from django.conf import settings
from django.contrib import admin, messages
from django.contrib.admin import ModelAdmin
from django.utils.translation import gettext_lazy as _

from django_seriously.authtoken.forms import TokenChangeForm
from django_seriously.authtoken.models import Token
from django_seriously.authtoken.utils import generate_token


class TokenAdmin(ModelAdmin):
    list_display = (
        "id",
        "user",
        "name",
        "scopes",
    )
    form = TokenChangeForm

    def save_model(self, request, obj: Token, form, change):
        if not change:
            token = generate_token()
            self.message_user(
                request=request,
                message=_(
                    'New bearer token is "{}". This can only be viewed once!'
                ).format(token.encoded_bearer),
                level=messages.WARNING,
            )
            obj.id = token.id
            obj.key = token.key
        return super().save_model(request, obj, form, change)

    def get_readonly_fields(self, request, obj=None):
        return "id", "created_at", "updated_at", "last_seen_at"


if "django_seriously.authtoken" in settings.INSTALLED_APPS:
    admin.site.register(Token, TokenAdmin)



import uuid

from django.db import models


class BaseModel(models.Model):
    """Opinionated Django base model"""

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        """
        Because the default is ridiculous. This guarantees that validation
        logic is executed in every non-bulk save situation. This comes at
        the expense of potentially running validation more than once.
        """
        self.full_clean()
        return super().save(*args, **kwargs)



# Generated by Django 3.2.9 on 2021-12-12 10:53

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Token',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(blank=True, max_length=25)),
                ('key', models.CharField(max_length=128, verbose_name='Key')),
                ('scopes', models.CharField(blank=True, help_text='comma-separated list of scopes. choices are .', max_length=50)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='auth_tokens', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



import base64
import os
import uuid

from django.contrib.auth.hashers import get_hasher

from django_seriously.settings import seriously_settings


class TokenContainer:
    """
    A token is comprised of two random 16 byte strings. Id serves as primary key on
    the token table for fast retrieval, while the second part contains the actual
    secret. The token table will only contain a salted and hashed value for
    comparison (key), while cleartext secret is never persisted.

    bearer = id + secret
    encoded_bearer = base64(bearer)
    key = PBKDF2(secret)
    """

    id: uuid.UUID
    key: str
    bearer: bytes
    encoded_bearer: str

    def __init__(self, id, key, bearer, encoded_bearer):
        self.id = id
        self.key = key
        self.bearer = bearer
        self.encoded_bearer = encoded_bearer


def generate_token() -> TokenContainer:
    token_id = uuid.uuid4()
    secret = os.urandom(16)
    raw_bearer_token = token_id.bytes + secret
    return TokenContainer(
        id=token_id,
        key=seriously_settings.MAKE_PASSWORD(secret),
        bearer=raw_bearer_token,
        encoded_bearer=base64.urlsafe_b64encode(raw_bearer_token).decode(),
    )


def make_password(password) -> str:
    """Default hasher function used by seriously_settings.MAKE_PASSWORD"""
    if not isinstance(password, (bytes, str)):
        raise TypeError(
            f"Password must be a string or bytes, got {type(password).__qualname__}."
        )
    hasher = get_hasher("pbkdf2_sha256")
    return hasher.encode(password, hasher.salt(), iterations=1_000)  # type: ignore


def check_password_rehash(raw_password: str) -> bool:
    return not raw_password.startswith("pbkdf2_sha256$1000$")



from datetime import datetime

import pytest
from django.core import exceptions
from django.utils import timezone
from pydantic import BaseModel as PydanticBaseModel

from django_seriously.utils.fields import ValidatedJSONField
from django_seriously.utils.models import BaseModel


class X(PydanticBaseModel):
    a: int
    b: str
    c: datetime


class ValidatedFieldTestModel(BaseModel):
    field = ValidatedJSONField(structure=X)
    field_list = ValidatedJSONField(structure=list[X])


X_EXAMPLE_DICT = {"a": 1, "b": "foo", "c": timezone.now()}


@pytest.mark.django_db
def test_validated_field():
    instance = ValidatedFieldTestModel.objects.create(
        field=X_EXAMPLE_DICT, field_list=[X_EXAMPLE_DICT]
    )
    assert isinstance(instance.field, dict)
    assert instance.field == X_EXAMPLE_DICT
    assert instance.field_list[0] == X_EXAMPLE_DICT


@pytest.mark.django_db
def test_validated_field_validation():
    with pytest.raises(exceptions.ValidationError):
        ValidatedFieldTestModel.objects.create(field=1, field_list=[X_EXAMPLE_DICT])
    with pytest.raises(exceptions.ValidationError):
        ValidatedFieldTestModel.objects.create(field=X_EXAMPLE_DICT, field_list=[1])
